// $Id$

//////////////////////////////////////////////////////////////////////////

#include "util.h"

#include <cv.h>
#include <highgui.h>
#include <imgptr.h>

//////////////////////////////////////////////////////////////////////////

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <iomanip>
#include <stdexcept>
#include <string>
#include <sstream>
#include <limits>

//////////////////////////////////////////////////////////////////////////

#ifndef TEST_LEVEL
#define TEST_LEVEL 1
#endif

#ifndef MIN_FSIZE
#define MIN_FSIZE 32
#endif

//////////////////////////////////////////////////////////////////////////

class prog_args
{
public:
  std::string in_name; // name of input file
  std::string out_name; // name of output file or directory

public:
  double color[3][2]; // color interval
  int dilate; // dilate the mask in given number of steps

public:
  prog_args(int argc, char * argv[]); // reads arguments
};

//////////////////////////////////////////////////////////////////////////

prog_args::prog_args(int argc, char * argv[])
{
  int const arg_num = argc - 1;
  int const arg_num_min = 2;
  int const arg_num_max = arg_num_min + 1;

  if(arg_num < arg_num_min || arg_num > arg_num_max)
  {
    std::ostringstream error;
    error
      << "arguments:"
      << " <in_name> <out_name>"
      << " [dilate]"
    ;
    throw std::runtime_error(error.str());
  }

  // MANDATORY ARGUMENTS

  in_name = argv[1]; // name of input file
  out_name = argv[2]; // name of output file or directory

  // OPTIONAL ARGUMENTS

  dilate =
    (arg_num >= arg_num_min + 1)
    ? atoi(argv[arg_num_min + 1])
    : 0; // do not dilate
}

//////////////////////////////////////////////////////////////////////////

class prog
{
private:
  prog_args const _args; // program arguments
  cv::img_ptr _in_img, _out_img; // input /output image

public:
  prog(int argc, char * argv[])
  :
    _args (argc, argv) // reads arguments
  {
    _init(); // initializes the program
    _run(); // executes the program
  }

private:
  prog(prog const &); // NOT IMPLEMENTED
  prog& operator=(prog const &);

private:
  void _init(); // initializes the program
  void _run(); // executes the program
};

//////////////////////////////////////////////////////////////////////////

void prog::_init()
{
  // OPEN INPUT

  _in_img = cvLoadImage(_args.in_name.c_str());

  if(_in_img == 0 || _in_img->imageData == 0)
  {
    std::ostringstream error;
    error << "couldn't load file: " << _args.in_name;
    throw std::runtime_error(error.str());
  }

  if(_in_img->nChannels != 3) // BGR
  {
    std::ostringstream error;
    error
      << "unsupported number of input channels: "
      << _in_img->nChannels;
    throw std::runtime_error(error.str());
  }

  if(_in_img->depth != (int) IPL_DEPTH_8U) // [0,255]
  {
    std::ostringstream error;
    error
      << "unsupported type of input data: "
      << ipl_depth_name(_in_img->depth);
    throw std::runtime_error(error.str());
  }

  if(_in_img->width < MIN_FSIZE || _in_img->height < MIN_FSIZE)
  {
    std::ostringstream error;
    error
      << "too small input image: "
      << _in_img->width << "x" << _in_img->height
      << " (must be >= " << MIN_FSIZE << ")";
    throw std::runtime_error(error.str());
  }

  CvSize const in_sz = cvSize(_in_img->width, _in_img->height);

#if TEST_LEVEL >= 1
  std::cerr
    << "Input: " << _args.in_name << " "
    << in_sz.width << "x" << in_sz.height << " "
    << std::endl;
#endif

  // OPEN OUTPUT

  CvSize const out_sz = in_sz; // output size == input size

#if TEST_LEVEL >= 1
  std::cerr
    << "Output: " << _args.out_name << " "
    << out_sz.width << "x" << out_sz.height << " "
    << std::endl;
#endif

  _out_img.reset(out_sz, CV_8UC1); // output image
}

//////////////////////////////////////////////////////////////////////////

void prog::_run()
{
  cvSet(_out_img, cvScalarAll(255));

  CvMemStorage * storage = cvCreateMemStorage();
  assert(storage != 0);

  cv::img_ptr in_gray (cvGetSize(_in_img), CV_8UC1);
  cvCvtColor(_in_img, in_gray, CV_BGR2GRAY);
  cvEqualizeHist(in_gray, in_gray);
  cvCanny(in_gray, in_gray, 100, 100);
  //cvSmooth(in_gray, in_gray, CV_GAUSSIAN, 3);
  //cvSmooth(in_gray, in_gray, CV_MEDIAN, 3);
  cvDilate(in_gray, in_gray, 0, 1);
  //cvCanny(in_gray, in_gray, 100, 100);
  //cvDilate(in_gray, in_gray, 0, 2);
  //cvErode(in_gray, in_gray, 0, 1);
  //cvCanny( src, dst, 50, 200, 3 );

  cvNamedWindow("edges");
  cvShowImage("edges", in_gray);
  cvWaitKey(-1);

  CvSize const sz = cvGetSize(_in_img);

  CvSeq * lines = cvHoughLines2(
    in_gray, storage, CV_HOUGH_STANDARD,
    1., CV_PI/180, 100
  );

  for(int i = 0; i < MIN(lines->total,100); ++i)
  {
    float* line = (float*)cvGetSeqElem(lines,i);
    float rho = line[0];
    float theta = line[1];
    if(
      (fabs(theta * 180 / CV_PI - 43) > 2) &&
      (fabs(theta * 180 / CV_PI - 133) > 2)
    ) continue;
    CvPoint pt1, pt2;
    double a = cos(theta), b = sin(theta);
    double x0 = a*rho, y0 = b*rho;
    pt1.x = cvRound(x0 + 1000*(-b));
    pt1.y = cvRound(y0 + 1000*(a));
    pt2.x = cvRound(x0 - 1000*(-b));
    pt2.y = cvRound(y0 - 1000*(a));
    cvLine(_in_img, pt1, pt2, CV_RGB(255,0,0), 3, 8 );
    cvLine(_out_img, pt1, pt2, CV_RGB(0,0,0), 3, 8 );
  }
  cvClearSeq(lines);

#if 0
  const double hough_param1 = 100, hough_param2 = 10;
  const int min_radius = 1, max_radius = 10;

  CvSeq * circles = cvHoughCircles(
      in_gray, storage, CV_HOUGH_GRADIENT, // applied on grayscale images
      1, 10, hough_param1, hough_param2, min_radius, max_radius
    );
  assert(circles != 0);

  for(int i = 0; i < circles->total; ++i)
  {
    const float * const p =
      reinterpret_cast<float *>(cvGetSeqElem(circles, i));

    int const x = cvRound(p[0]);
    int const y = cvRound(p[1]);
    int const r = cvRound(p[2]);

    cvCircle(
      _in_img, cvPoint(x, y), r,
      CV_RGB(0, 0, 255), 2, 8, 0
    );
  }
  cvClearSeq(circles);
#endif

  cvNamedWindow("image");
  cvShowImage("image", _in_img);
  cvWaitKey(-1);

  cvReleaseMemStorage(&storage);

  if(_args.dilate > 0)
  {
    cvDilate(_out_img, _out_img, 0, _args.dilate);
    cvSmooth(_out_img, _out_img, CV_MEDIAN);
  }
  else if(_args.dilate < 0)
  {
    cvErode(_out_img, _out_img, 0, _args.dilate);
    cvSmooth(_out_img, _out_img, CV_MEDIAN);
  }

  cvSaveImage(_args.out_name.c_str(), _out_img);
}

//////////////////////////////////////////////////////////////////////////

int main(int argc, char * argv[])
{
  std::string const prog_name = argv[0];
  try
  {
    prog(argc, argv);
    return EXIT_SUCCESS;
  }
  catch(std::exception const & e)
  {
    std::cerr
      << prog_name << ": " << e.what()
      << std::endl;
  }
  return EXIT_FAILURE;
}

//////////////////////////////////////////////////////////////////////////

// End of File
