#ifndef IMGPTR_H
#define IMGPTR_H

// Copyleft (x) Sandor Fazekas, 2008. All rights reversed.
// $Id: imgptr.h,v 1.3 2008/07/08 07:10:28 fazekas Exp $

//////////////////////////////////////////////////////////////////////////

#ifndef TEST_LEVEL
#define TEST_LEVEL 0
#endif

#ifndef MIN_TEST_LEVEL
#define MIN_TEST_LEVEL 2
#endif

//////////////////////////////////////////////////////////////////////////

#include <cv.h>
#include <cassert>

#if TEST_LEVEL >= MIN_TEST_LEVEL
#include <iostream>
#endif

//////////////////////////////////////////////////////////////////////////
namespace cv {
//////////////////////////////////////////////////////////////////////////

class img_ptr
{
public:
  typedef IplImage * data_ptr;
  typedef IplImage const * data_const_ptr;

private:
  data_ptr _p; // pointer to OWNED data

public:
  img_ptr()
  :
    _p (0)
  {
    // NO DATA
  }

  img_ptr(CvSize size, int depth, int channels)
  :
    _p (0)
  {
    reset(size, depth, channels); // ALLOCATES DATA
  }

  img_ptr(CvSize size, int type)
  :
    _p (0)
  {
    reset(size, type); // ALLOCATES DATA
  }

  explicit
  img_ptr(CvArr const * d)
  :
    _p (0)
  {
    reset(d); // CLONES DATA, if there is any
  }

  explicit
  img_ptr(CvMat const & mat)
  :
    _p (0)
  {
    reset(mat); // CLONES DATA
  }

  explicit
  img_ptr(IplImage const & img)
  :
    _p (0)
  {
    reset(img); // CLONES DATA
  }

  ~img_ptr()
  {
    reset(); // DROPS DATA
  }

public:
  img_ptr(img_ptr const & a)
  :
    _p (0)
  {
  #if TEST_LEVEL >= MIN_TEST_LEVEL
    if(a._p != 0)
    {
      std::cout
        << "warning: img_ptr COPY CONSTRUCTOR called"
        << std::endl;
    }
  #endif
    reset(a._p); // CLONES DATA
  }

  img_ptr & operator=(img_ptr const & a)
  {
    reset(a._p); // CLONES DATA, if there was any
    return *this;
  }

  img_ptr & operator=(CvArr const * d)
  {
    reset(d); // CLONES DATA, if there was any
    return *this;
  }

  img_ptr & operator=(CvMat const & mat)
  {
    reset(mat); // CLONES DATA, if there was any
    return *this;
  }

  img_ptr & operator=(IplImage const & img)
  {
    reset(img); // CLONES DATA, if there was any
    return *this;
  }

public:
  void reset() // DROPS DATA, if there was any
  {
    if(_p != 0)
    {
    /////////////////////////////////////////////
      cvReleaseImage(&_p); // DEALLCATES memory
    /////////////////////////////////////////////
      _p = 0; // NO DATA
    }
  }

  void reset(CvSize size, int depth, int channels)
  {
    reset(); // DROPS DATA, if there was any
  //////////////////////////////////////////////////////////////////
    _p = cvCreateImage(size, depth, channels); // ALLOCATES memory
  //////////////////////////////////////////////////////////////////
    assert(_p != 0);
  }

  void reset(CvSize size, int type) // ALLOCATES DATA
  {
    reset(size, cvCvToIplDepth(type), CV_MAT_CN(type));
  }

  void reset(CvArr const * d) // CLONES DATA
  {
    if(d == 0)
    {
      reset(); // DROPS DATA, if there was any
      return; // NOTHING TO CLONE
    }

    assert(CV_IS_IMAGE(d) || CV_IS_MAT(d));

    // NOTE: we try to avoid problems caused by aliasing

    IplImage hdr; // temporary image header
    IplImage * img = cvGetImage(d, &hdr);
    if(_p != 0 && _p->imageData == img->imageData)
    {
      return; // NOTHING TO CLONE, this the same data
    }

    img_ptr img_clone; // stores a clone
  ////////////////////////////////////////////////////
    img_clone._p = cvCloneImage(img); // CLONES DATA
  ////////////////////////////////////////////////////
    reset(); // DEALLOCATES OLD DATA, if there was any
    _p = img_clone._p; // stores reference to NEW DATA
    img_clone._p = 0; // RELEASES ownership
    assert(_p != 0);
  }

  void reset(CvMat const & mat)
  {
    reset(&mat); // CLONES DATA
  }

  void reset(IplImage const & img)
  {
    reset(&img); // CLONES DATA
  }

public:
  operator data_ptr() { return _p; }; // ACCESS to owned data
  operator data_const_ptr() const { return _p; };

  data_ptr operator->() { return _p; }; // ACCESS to data members
  data_const_ptr operator->() const { return _p; };

private:
  void * _get_row(int y) const
  {
    if(_p == 0) return 0; // NO DATA
    assert(0 <= y && y < _p->height);
    return _p->imageData + y * _p->widthStep;
  }

public:
  void * row(int y) { return _get_row(y); }
  void const * row(int y) const { return _get_row(y); }

public:
  friend void swap(img_ptr & a, img_ptr & b)
  {
    if(&a == &b) return; // nothing to exchange
    data_ptr p = a._p;
    a._p = b._p;
    b._p = p;
  }
};

//////////////////////////////////////////////////////////////////////////
} // namespace cv
//////////////////////////////////////////////////////////////////////////

#endif // IMGPTR_H
