// $Id$

//////////////////////////////////////////////////////////////////////////

#include "inpaint.h"
#include "inpaint_dt.h"
#include "util.h"

#include <cv.h>
#include <highgui.h>
#include <imgptr.h>

//////////////////////////////////////////////////////////////////////////

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <iomanip>
#include <stdexcept>
#include <string>
#include <sstream>
#include <limits>

//////////////////////////////////////////////////////////////////////////

#ifndef TEST_LEVEL
#define TEST_LEVEL 1
#endif

#ifndef MIN_FSIZE
#define MIN_FSIZE 32
#endif

//////////////////////////////////////////////////////////////////////////

class prog_args
{
public:
  std::string in_name; // name of input file
  std::string mask_name; // mask of 'in_name'
  std::string out_name; // name of output file or directory

public:
  int it_num; // number of iteration steps
  int lev_num; // number of resoltion levels
  double lambda; // scaling factor
  double alpha; // diffusion parameter
  int dilate; // dilate or erode the mask given number of times

public:
  prog_args(int argc, char * argv[]); // reads arguments
};

//////////////////////////////////////////////////////////////////////////

prog_args::prog_args(int argc, char * argv[])
{
  int const arg_num = argc - 1;
  int const arg_num_min = 7;
  int const arg_num_max = arg_num_min + 1;

  if(arg_num < arg_num_min || arg_num > arg_num_max)
  {
    std::ostringstream error;
    error
      << "arguments:"
      << " <in_name> <mask_name> <out_name>"
      << " <it_num> <lev_num> <lambda> <alpha>"
    ;
    throw std::runtime_error(error.str());
  }

  // MANDATORY ARGUMENTS

  in_name = argv[1]; // name of input file
  mask_name = argv[2]; // mask of input file
  out_name = argv[3]; // name of output file or directory

  it_num = atoi(argv[4]); // number of iteration steps

  if(it_num < 0)
  {
    std::ostringstream error;
    error << "invalid number of iterations: " << it_num;
    throw std::runtime_error(error.str());
  }

  lev_num = atoi(argv[5]); // number of iteration steps

  if(lev_num < 0)
  {
    std::ostringstream error;
    error << "invalid number of resoltion levels: " << lev_num;
    throw std::runtime_error(error.str());
  }

  lambda = atof(argv[6]); // scaling factor

  if(lambda <= 0 || lambda >= 1)
  {
    std::ostringstream error;
    error
      << "invalid scaling factor (lambda): "
      << lambda << " (should be > 0 and < 1)";
    throw std::runtime_error(error.str());
  }

  alpha = atof(argv[7]); // diffusion parameter

  if(alpha < 0)
  {
    std::ostringstream error;
    error
      << "invalid diffusion parameter (alpha): "
      << alpha << " (should be > 0)";
    throw std::runtime_error(error.str());
  }

  // OPTIONAL ARGUMENTS

  dilate =
    (arg_num >= arg_num_min + 1)
    ? atoi(argv[arg_num_min + 1])
    : 0; // no dilation
}

//////////////////////////////////////////////////////////////////////////

class prog
{
private:
  prog_args const _args; // program arguments
  inpaint_dt_func _inpaint_dt; // based on distance transform
  mm_inpaint_func _inpaint_anis; // based on anisotropic diffusion
  cv::img_ptr _in_img, _mask_img, _out_img; // input/mask/output image

public:
  prog(int argc, char * argv[])
  :
    _args (argc, argv) // reads arguments
  {
    _init(); // initializes the program
    _run(); // executes the program
  }

private:
  prog(prog const &); // NOT IMPLEMENTED
  prog& operator=(prog const &);

private:
  void _init(); // initializes the program
  void _run(); // executes the program
};

//////////////////////////////////////////////////////////////////////////

void prog::_init()
{
  // OPEN INPUT

  _in_img = cvLoadImage(_args.in_name.c_str());

  if(_in_img == 0 || _in_img->imageData == 0)
  {
    std::ostringstream error;
    error << "couldn't load file: " << _args.in_name;
    throw std::runtime_error(error.str());
  }

  if(_in_img->nChannels != 3) // BGR
  {
    std::ostringstream error;
    error
      << "unsupported number of input channels: "
      << _in_img->nChannels;
    throw std::runtime_error(error.str());
  }

  if(_in_img->depth != (int) IPL_DEPTH_8U) // [0,255]
  {
    std::ostringstream error;
    error
      << "unsupported type of input data: "
      << ipl_depth_name(_in_img->depth);
    throw std::runtime_error(error.str());
  }

  if(_in_img->width < MIN_FSIZE || _in_img->height < MIN_FSIZE)
  {
    std::ostringstream error;
    error
      << "too small input image: "
      << _in_img->width << "x" << _in_img->height
      << " (must be >= " << MIN_FSIZE << ")";
    throw std::runtime_error(error.str());
  }

  CvSize const in_sz = cvSize(_in_img->width, _in_img->height);

#if TEST_LEVEL >= 1
  std::cerr
    << "Input: " << _args.in_name << " "
    << in_sz.width << "x" << in_sz.height << " "
    << std::endl;
#endif

  {
    int w_k = in_sz.width, h_k = in_sz.height;
    for(int k = 1; k < _args.lev_num; ++k)
    {
      w_k = static_cast<int>(floor(w_k * _args.lambda));
      h_k = static_cast<int>(floor(h_k * _args.lambda));
      if(w_k < MIN_FSIZE || h_k < MIN_FSIZE)
      {
        std::ostringstream error;
        error
          << "too many levels: " << _args.lev_num
          << " (should be <= " << k << ")";
        throw std::runtime_error(error.str());
      }
    }
  }

  // READ MASK

  _mask_img = cvLoadImage(
    _args.mask_name.c_str(), CV_LOAD_IMAGE_GRAYSCALE
  );

  if(_mask_img == 0 || _mask_img->imageData == 0)
  {
    std::ostringstream error;
    error << "couldn't load mask: " << _args.mask_name;
    throw std::runtime_error(error.str());
  }

  assert(_mask_img->nChannels == 1);

  if(_mask_img->depth != (int) IPL_DEPTH_8U) // [0,255]
  {
    std::ostringstream error;
    error
      << "unsupported type of mask data: "
      << ipl_depth_name(_mask_img->depth);
    throw std::runtime_error(error.str());
  }

  if(cvGetSize(_mask_img) != in_sz)
  {
    std::ostringstream error;
    error
      << "the size of mask and image should match: "
      << _mask_img->width << "x" << _mask_img->height
      << " (must be "
      << _in_img->width << "x" << _in_img->height << ")";
    throw std::runtime_error(error.str());
  }

  cvThreshold(_mask_img, _mask_img, 128, 255, CV_THRESH_BINARY);
  cvNot(_mask_img, _mask_img);

  if(_args.dilate > 0)
  {
    int const dilate = _args.dilate;
    for(int step = 0; step < dilate; ++step)
    {
      cvDilate(_mask_img, _mask_img);
    }
  }
  else if(_args.dilate < 0)
  {
    int const erode = - _args.dilate;
    for(int step = 0; step < erode; ++step)
    {
      cvErode(_mask_img, _mask_img);
    }
  }

  // OPEN OUTPUT

  CvSize const out_sz = in_sz; // output size == input size

#if TEST_LEVEL >= 1
  std::cerr
    << "Output: " << _args.out_name << " "
    << out_sz.width << "x" << out_sz.height << " "
    << std::endl;
#endif

  _out_img.reset(out_sz, CV_8UC3); // output image
}

//////////////////////////////////////////////////////////////////////////

void prog::_run()
{
#if TEST_LEVEL >= 3
  cvNamedWindow("input");
  cvShowImage("input", _in_img);

  cvNamedWindow("mask");
  cvShowImage("mask", _mask_img);

  std::cerr << "Press any key to start!" << std::endl;
  cvWaitKey(-1); // WAITS FOR ANY KEY
  std::cerr << "Running ..." << std::endl;
#endif

#if 0
#if 0 // HACK
#if 1 // HACK
  cvSet(_in_img, cvScalarAll(0), _mask_img);
#endif

  _inpaint_dt(_in_img, _mask_img, _out_img);

#if 1 // HACK
  cv::img_ptr out (cvGetSize(_out_img), CV_8UC3);
  cvSmooth(_out_img, out, CV_MEDIAN, 11);
  cvCopy(out, _out_img, _mask_img);
#endif

  cvCopy(_out_img, _in_img);
#endif
#else
#if 0 // HACK
  cvSet(_in_img, cvScalarAll(0), _mask_img);
#endif

  #if 0 // HACK
    cvNamedWindow("original mask");
    cvShowImage("original mask", _mask_img);
    cvWaitKey(-10);
  #endif

  int const it_num = 10;
  for(int step = 0; step < it_num; ++step)
  {
    _inpaint_dt(_in_img, _mask_img, _out_img);

  #if 1 // HACK
    cv::img_ptr out (cvGetSize(_out_img), CV_8UC3);
    cvSmooth(_out_img, out, CV_MEDIAN, 11);
    cvCopy(out, _out_img, _mask_img);
  #endif

  #if 0 // HACK
    cvNamedWindow("out");
    cvShowImage("out", _out_img);
  #endif

    cv::img_ptr diff_img (cvGetSize(_out_img), CV_8UC3);
    cvAbsDiff(_out_img, _in_img, diff_img);

    cv::img_ptr diff_img_ch[3];
    for(int a = 0; a < 3; ++a)
    {
      diff_img_ch[a].reset(cvGetSize(_out_img), CV_8UC1);
    }

    cvSplit(diff_img, diff_img_ch[0], diff_img_ch[1], diff_img_ch[2], 0);

    double const thresh = 1 + 10 * (it_num - step) / (double) it_num;
    cv::img_ptr new_mask (cvGetSize(_mask_img), CV_8UC1);
    cvSetZero(new_mask);
    for(int a = 0; a < 3; ++a)
    {
    #if 1
      cvThreshold(
        diff_img_ch[a], diff_img_ch[a],
        thresh, 255, CV_THRESH_BINARY
      );
    #else
      cvAdaptiveThreshold(
        diff_img_ch[a], diff_img_ch[a],
        255, CV_ADAPTIVE_THRESH_MEAN_C,
        CV_THRESH_BINARY, 11, 10
      );
    #endif
      cvOr(new_mask, diff_img_ch[a], new_mask);
    }
    //cvSmooth(new_mask, new_mask, CV_MEDIAN, 3);
    //cvDilate(new_mask, new_mask);
    cvAnd(_mask_img, new_mask, _mask_img);
    //cvCopy(new_mask, _mask_img);

  #if 1 // HACK
    cvNamedWindow("mask");
    cvShowImage("mask", _mask_img);
    cvWaitKey(10);
  #endif
  }
#endif

#if TEST_LEVEL >= 3
  cvNamedWindow("input");
  cvShowImage("input", _in_img);
  cvWaitKey(10);
#endif

  cv::img_ptr in_img (cvGetSize(_in_img), CV_32FC3);
  cv::img_ptr out_img (cvGetSize(_out_img), CV_32FC3);

  cvConvert(_in_img, in_img);
  cart_to_spheric(in_img, in_img);

  _inpaint_anis(
    in_img, _mask_img, out_img, _args.it_num,
    _args.lev_num, _args.lambda, _args.alpha
  );

#if 1 // HACK
  cv::img_ptr texture (cvGetSize(in_img), CV_32FC3);

  cvSmooth(in_img, texture, CV_GAUSSIAN, 3);
  cvSub(in_img, texture, texture);

#if 1 // HACK
  _inpaint_dt(texture, _mask_img, texture);
#endif

  cvAdd(out_img, texture, out_img, _mask_img);
#endif

  spheric_to_cart(out_img, out_img);
  cvConvert(out_img, _out_img);

#if 0 // HACK
  cv::img_ptr out_img_ (_out_img);
  cvSmooth(_out_img, out_img_, CV_MEDIAN, 3);
  cvDilate(_mask_img, _mask_img, 0, 1);
  cvCopy(out_img_, _out_img, _mask_img);
#endif

  cvSaveImage(_args.out_name.c_str(), _out_img);

#if TEST_LEVEL >= 3
  cvNamedWindow("result");
  cvShowImage("result", _out_img);
  std::cerr << "Press any key ..." << std::endl;
  cvWaitKey(-1);
#endif
}

//////////////////////////////////////////////////////////////////////////

int main(int argc, char * argv[])
{
  std::string const prog_name = argv[0];
  try
  {
    prog(argc, argv);
    return EXIT_SUCCESS;
  }
  catch(std::exception const & e)
  {
    std::cerr
      << prog_name << ": " << e.what()
      << std::endl;
  }
  return EXIT_FAILURE;
}

//////////////////////////////////////////////////////////////////////////

// End of File
