// $Id$

//////////////////////////////////////////////////////////////////////////

#include "inpaint_dt.h"

//////////////////////////////////////////////////////////////////////////

#ifndef TEST_LEVEL
#define TEST_LEVEL 0
#endif

#ifndef MIN_TEST_LEVEL
#define MIN_TEST_LEVEL 2
#endif

//////////////////////////////////////////////////////////////////////////

#ifndef MIN_FSIZE
#define MIN_FSIZE 32
#endif

//////////////////////////////////////////////////////////////////////////

#include <cv.h>
#include <imgptr.h>
#include <unwarp.h>

//////////////////////////////////////////////////////////////////////////

#include <cstdlib>
#include <cassert>
#include <cmath>
#include <vector>
#include <new>

//////////////////////////////////////////////////////////////////////////

#if TEST_LEVEL >= MIN_TEST_LEVEL
#include <highgui.h>
#include <iostream>
#include <sstream>
#endif

//////////////////////////////////////////////////////////////////////////
namespace _inpaint_dt_ {
//////////////////////////////////////////////////////////////////////////

inline
bool operator==(CvSize const & a, CvSize const & b)
{
  return (a.width == b.width) && (a.height == b.height);
}

inline
bool operator!=(CvSize const & a, CvSize const & b)
{
  return !(a == b);
}

//////////////////////////////////////////////////////////////////////////
} // namespace _inpaint_dt_
//////////////////////////////////////////////////////////////////////////

using namespace _inpaint_dt_;

//////////////////////////////////////////////////////////////////////////

class _dt_inp // simple inpainting based on distance transform
{
protected:
  CvSize _img_size; // size of image
  cv::img_ptr _mask_dt; // distance transform of mask
  cv::img_ptr _dt_grad[2]; // gradient of distance transform
  cv::img_ptr _displ[2]; // inpainting displacement

public:
  _dt_inp()
  :
    _img_size (cvSize(0,0))
  {
  }

public:
  void inpaint(CvArr const * in, CvArr const * mask, CvArr * out);
};

//////////////////////////////////////////////////////////////////////////

void _dt_inp::inpaint(CvArr const * in, CvArr const * mask, CvArr * out)
{
  CvSize const sz = cvGetSize(in);
  assert(cvGetSize(mask) == sz);
  assert(cvGetSize(out) == sz);

  assert(cvGetElemType(mask) == CV_8UC1);

  if(sz != _img_size)
  {
    _mask_dt.reset(sz, CV_32FC1);
    _dt_grad[0].reset(sz, CV_32FC1);
    _dt_grad[1].reset(sz, CV_32FC1);
    _displ[0].reset(sz, CV_32SC1);
    _displ[1].reset(sz, CV_32SC1);
    _img_size = sz;
  }

  cvDistTransform(mask, _mask_dt);

#if 0 // HACK
  cvSmooth(_mask_dt, _mask_dt, CV_GAUSSIAN, 3);
#endif

#if TEST_LEVEL >= 4
  {
    double max_dist = 0;
    cvMinMaxLoc(_mask_dt, 0, &max_dist);
    cv::img_ptr mask_dist (sz, CV_32FC1);
    cvScale(_mask_dt, mask_dist, 1.0 / max_dist);
    cvNamedWindow("distance transform");
    cvShowImage("distance transform", mask_dist);
    std::cerr << "Press any key to continue!" << std::endl;
    cvWaitKey(-1); // WAITS FOR ANY KEY
  }
#endif

#if 0 // HACK
#if 0
  int const der_kern = CV_SCHARR;
  double const der_scale = 1.0 / 32.0;
#else
  int const der_kern = 3; // Sobel
  double const der_scale = 1.0 / 8.0;
#endif
#else
  int const der_kern = 1; // kernel: -1 0 1
  double const der_scale = 1.0 / 2.0;
#endif

  cvSobel(_mask_dt, _dt_grad[0], 1, 0, der_kern);
  cvSobel(_mask_dt, _dt_grad[1], 0, 1, der_kern);

  cvScale(_dt_grad[0], _dt_grad[0], der_scale);
  cvScale(_dt_grad[1], _dt_grad[1], der_scale);

#if 0 // HACK
  cvSmooth(_dt_grad[0], _dt_grad[0], CV_GAUSSIAN, 3);
  cvSmooth(_dt_grad[1], _dt_grad[1], CV_GAUSSIAN, 3);
#endif

  for(int y = 0; y < sz.height; ++y)
  {
    float const * const dist_row =
      static_cast<float *>(_mask_dt.row(y));

    float * const grad_row[2] =
    {
      static_cast<float *>(_dt_grad[0].row(y)),
      static_cast<float *>(_dt_grad[1].row(y)),
    };

    int * const displ_row[2] =
    {
      static_cast<int *>(_displ[0].row(y)),
      static_cast<int *>(_displ[1].row(y)),
    };

    for(int x = 0; x < sz.width; ++x)
    {
      if(dist_row[x] == 0)
      {
        displ_row[0][x] = displ_row[1][x] = 0;
        continue;
      }

      float const gr[2] = { grad_row[0][x], grad_row[1][x] };

      float const gr_thr = 1e-8; // THRESHOLD
      float const gr_nr = sqrt(gr[0] * gr[0] + gr[1] * gr[1]);

      if(gr_nr < gr_thr)
      {
        displ_row[0][x] = displ_row[1][x] = 0;
        // TODO: handle this case in a more clever way
        continue;
      }

      float const nr[2] = { - gr[0] / gr_nr, - gr[1] / gr_nr };

      float displ = dist_row[x];
      while(true)
      {
        float const dr[2] = { displ * nr[0], displ * nr[1] };

        int const x_ = static_cast<int>(floor(x + dr[0]));
        if(x_ < 0 || x_ >= sz.width) break;

        int const y_ = static_cast<int>(floor(y + dr[1]));
        if(y_ < 0 || y_ >= sz.height) break;

        float const * const dist_row_ =
          static_cast<float *>(_mask_dt.row(y_));
        if(dist_row_[x_] == 0) break;

        displ += dist_row_[x_];
      }

      displ_row[0][x] = static_cast<int>(floor(displ * nr[0]));
      displ_row[1][x] = static_cast<int>(floor(displ * nr[1]));
    }
  }

#if TEST_LEVEL >= 4
  for(int a = 0; a < 2; ++a)
  {
    double max_displ = 0;
    cvMinMaxLoc(_displ[a], 0, &max_displ);
    cv::img_ptr displ (sz, CV_32FC1);
    cvScale(_displ[a], displ, 1.0 / max_displ);
    cvNamedWindow("displacement");
    cvShowImage("displacement", displ);
    std::cerr << "Press any key to continue!" << std::endl;
    cvWaitKey(-1); // WAITS FOR ANY KEY
  }
#endif

  cvExUnwarp(in, out, _displ[0], _displ[1]);
}

//////////////////////////////////////////////////////////////////////////
// EXPORT
//////////////////////////////////////////////////////////////////////////

inpaint_dt_func::inpaint_dt_func()
:
  _intern (0)
{
  // NO MEMORY ALLOCATED
}

//////////////////////////////////////////////////////////////////////////

inpaint_dt_func::~inpaint_dt_func()
{
  reset(); // DEALLOCATES MEMORY, if there was any allocated
}

//////////////////////////////////////////////////////////////////////////

inpaint_dt_func::inpaint_dt_func(
  inpaint_dt_func const & func
)
:
  _intern (0)
{
  // there is NO REASON TO COPY anything
  // NO MEMORY ALLOCATED
}

//////////////////////////////////////////////////////////////////////////

inpaint_dt_func &
  inpaint_dt_func::operator=(inpaint_dt_func const & func)
{
  // there is NO REASON TO COPY anything
  return *this;
}

//////////////////////////////////////////////////////////////////////////

void inpaint_dt_func::reset()
{
  if(_intern == 0) return; // IGNORED
  delete static_cast<_dt_inp *>(_intern);
  _intern = 0;
}

//////////////////////////////////////////////////////////////////////////

void inpaint_dt_func::operator()(
  CvArr const * in, CvArr const * mask, CvArr * out
)
{
  if(_intern == 0)
  {
    _intern = new _dt_inp ();
    assert(_intern != 0);
  }

  _dt_inp * intern = static_cast<_dt_inp *>(_intern);

  intern->inpaint(in, mask, out);
}

//////////////////////////////////////////////////////////////////////////

// End of File
